@@ -1,7 +1,4 @@
 {-# LANGUAGE CPP #-}
-#if __GLASGOW_HASKELL__ >= 704
-{-# LANGUAGE Safe #-}
-#endif
 {- |
 Module      :  System.FilePath.ByteString
 Copyright   :  (c) Neil Mitchell 2005-2014, (c) Joey Hess 2019
